/**
 * Copyright (C) 2011 Jacob Scott <jascottytechie@gmail.com>
 * Description: Global Shop System for Minecraft
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package me.jascotty2.bettershop;

import java.io.IOException;
import java.sql.SQLException;
import me.jascotty2.bettershop.shop.Shop;
import me.jascotty2.bettershop.shop.BSTransactionLog;
import me.jascotty2.bettershop.shop.BSItemStock;
import me.jascotty2.bettershop.shop.BSPriceList;
import me.jascotty2.bettershop.spout.SpoutKeyListener;
import me.jascotty2.bettershop.spout.SpoutPopupListener;
import me.jascotty2.bettershop.utils.BetterShopLogger;
import me.jascotty2.bettershop.commands.BSCommandManager;
import me.jascotty2.bettershop.commands.HelpCommands;
import me.jascotty2.bettershop.regionshops.RegionShopManager;
import me.jascotty2.bettershop.signshop.BSSignShop;
import me.jascotty2.bettershop.chestshop.BSChestShop;

import org.bukkit.command.Command;
import org.bukkit.command.CommandSender;
import org.bukkit.entity.Player;
import org.bukkit.plugin.PluginDescriptionFile;
import org.bukkit.plugin.java.JavaPlugin;
import org.bukkit.plugin.Plugin;
import org.bukkit.plugin.PluginManager;
import org.bukkit.Server;
import org.bukkit.ChatColor;
import org.bukkit.Location;

import java.util.Timer;
import java.util.TimerTask;
import java.util.Collection;

import me.jascotty2.bettershop.enums.ShopMethod;
import me.jascotty2.lib.bukkit.item.JItemDB;
import me.jascotty2.lib.bukkit.item.CreatureItem.EntityListen;
import me.jascotty2.lib.bukkit.commands.CommandException;
import me.jascotty2.lib.bukkit.commands.CommandPermissionsException;
import me.jascotty2.lib.bukkit.commands.CommandUsageException;
import me.jascotty2.lib.bukkit.commands.MissingNestedCommandException;
import me.jascotty2.lib.bukkit.commands.WrappedCommandException;
import me.jascotty2.lib.util.ArrayManip;
import me.jascotty2.lib.util.Str;
import org.mcstats.Metrics;

/**
 * BetterShop for Bukkit
 */
public class BetterShop extends JavaPlugin {

	protected static Plugin bettershopPlugin = null;
	protected final static BSConfig config = new BSConfig();
	protected final static RegionShopManager shopManager = new RegionShopManager();
	protected static BSSignShop signShop = null;
	protected static BSChestShop chestShop = null;
	protected static BSEcon economy;
	protected static TransactionHandler shopHandler;
	private BSPluginListener pListener = null;
	// for animal/monster purchases
	public final EntityListen entityListener = new EntityListen();
	private final static BSCommandManager commandManager = new BSCommandManager();
	protected static String lastCommand = "";
	// for spout-related classes
	protected static SpoutKeyListener keyListener = null;
	protected static SpoutPopupListener buttonListener = null;

	@Override
	public void onEnable() {
		bettershopPlugin = this;
		PluginDescriptionFile pdfFile = this.getDescription();
		BetterShopLogger.init(this.getLogger());
		BetterShopLogger.setUsePrefix(false);
		
		config.extractDefaults();
		// ready items db (needed for pricelist, sorting in config, item lookup,
		// ...)
		try {
			JItemDB.load(BSConfig.itemDBFile);
			// Log("Itemsdb loaded");
		} catch (Exception e) {
			BetterShopLogger.Severe("cannot load items db: closing plugin", e);
			this.setEnabled(false);
			return;
		}

		config.load();
		
		economy = new BSEcon(this);
		shopHandler = new TransactionHandler(economy, shopManager);
		
		if (shopManager.load() > 0) {
			BetterShopLogger.Severe("Error while enabling Shop");
		}
		signShop = new BSSignShop(this);

		if (config.signShopEnabled && !signShop.load()) {
			BetterShopLogger.Severe("cannot load sign shop database");
		}
		if (config.signShopEnabled && config.signWEprotection) {
			signShop.startProtecting();
		}

		signShop.registerEvents();


		chestShop = new BSChestShop(this);

		if (config.chestShopEnabled && !chestShop.load()) {
			BetterShopLogger.Severe("cannot load chest shop database");
		}

		chestShop.registerEvents();


		// for monster purchasing
		PluginManager pm = getServer().getPluginManager();
		pm.registerEvents(entityListener, this);

		// monitor plugins - if any are enabled/disabled by a plugin manager
		pListener = new BSPluginListener(this);
		pm.registerEvents(pListener, this);

		BetterShopLogger.Info(pdfFile.getName() + " version "
				+ pdfFile.getVersion() + " is enabled!");

		try {
			Metrics metrics = new Metrics(this);
			metrics.start();
		} catch (IOException e) {
			// Failed to submit the stats :-(
		}
	}

	@Override
	public void onDisable() {
		// NOTE: All registered events are automatically unregistered when a plugin is disabled
		try {
			lastCommand = "(disabling)";

			if (signShop != null) {
				signShop.save();
				signShop.stopProtecting();
			}
			signShop = null;

			if (chestShop != null) {
				chestShop.save();
				chestShop.closeAllChests();
			}
			chestShop = null;

			shopManager.closeAll();

			keyListener = null;
			buttonListener = null;

			BetterShopLogger.CloseCommandLog();

			BetterShopLogger.Fine("disabled");

		} catch (Throwable t) {
			BetterShopLogger.Severe("error disabling..", t);
		}
	}

	@Override
	public boolean onCommand(CommandSender sender, Command command,
			String commandLabel, String[] args) {
		String commandName = command.getName().toLowerCase(), argStr = Str.concatStr(args, " ");
		if (!commandManager.hasCommand(commandName)) {
			// No command found!
			return false;
		}
		lastCommand = (sender instanceof Player ? "player:" : "console:") + commandName + " " + argStr;

		if (sender instanceof Player
				&& (!shopManager.locationHasShop(((Player) sender).getLocation())
				&& (Str.isIn(commandName,
				new String[]{"shopbuy", "shopbuyall", "shopbuystack",
					"shopsell", "shopsellall", "shopsellstack",
				 "shoplist", "shopitems", "shopcheck", "shoplistkits"
				 /*, "shopadd",
				 * "shopremove"
				 */})
				&& !commandShopEnabled(((Player) sender).getLocation())))) {
			BSutils.sendMessage(sender, config.getString("regionShopDisabled"));
			return true;
		}

		if (config.logCommands) {
			BetterShopLogger.LogCommand(
					sender instanceof Player ? ((Player) sender).getName() : "(console)",
					commandLabel + " " + argStr);
		}

		if (!BSEcon.active()
				&& (commandName.contains("buy")
				|| commandName.contains("sell")
				|| argStr.contains("buy") || argStr.contains("sell"))) {
			BSutils.sendMessage(sender, ChatColor.RED
					+ "BetterShop is missing a dependency. Check the console.");
			BetterShopLogger.Severe("Missing: an economy plugin\nsupported: iConomy 4,5,6, BOSEcon 4,5, MultiCurrency, EssensialsEco");
			return true;
		}

		checkRestock();

		try {
			commandManager.execute(sender, commandName, args);
			return true;
		} catch (CommandPermissionsException e) {
			BSutils.sendMessage(sender,
					BetterShop.getSettings().getString("permdeny").replace("<perm>", e.getNode()));
			return true;
		} catch (MissingNestedCommandException e) {
			//BSutils.sendMessage(sender, ChatColor.RED + e.getMessage());
			BSutils.sendMessage(sender, ChatColor.RED + e.getUsage());
		} catch (CommandUsageException e) {
			BSutils.sendMessage(sender, ChatColor.RED + e.getMessage());
			BSutils.sendMessage(sender, ChatColor.RED + e.getUsage());
		} catch (WrappedCommandException e) {
			Throwable t = e.getCause();
			if(t instanceof SQLException){
				BetterShopLogger.Severe("Problem with MySQL Connection", t);
			} else if(t instanceof IOException){
				BetterShopLogger.Severe("Problem with File Access", t);
			} else {
				BetterShopLogger.Severe("Error executing a command", t);
			}
			BSutils.sendMessage(sender, ChatColor.RED + "Problem Executing Command!");
			return true;
		} catch (CommandException e) {
			//BetterShopLogger.Warning(e);
			BSutils.sendMessage(sender, ChatColor.RED + /*"Problem Executing Command!"*/ e.getMessage());
		} catch (Exception e) {
			BetterShopLogger.Severe("Unexpected Error executing a command", e);
			BSutils.sendMessage(sender, ChatColor.RED + "Unexpected Error executing command");
			return true;
		} catch (Throwable t) {
			BetterShopLogger.Severe("Unexpected Error in BetterShop", t);
			BSutils.sendMessage(sender, ChatColor.RED + "Unexpected Error executing command");
			return true;
		}
		//TODO: help can get help from any command.
		//HelpCommands.help(sender, ArrayManip.arrayConcat(new String[]{command.getName()}, args));
		return true;
	}

	public static void setLastCommand(String lastCommand) {
		StackTraceElement stm[] = (new Exception()).getStackTrace();
		if (stm[1].toString().startsWith(BetterShop.class.getPackage().getName())) {
			BetterShop.lastCommand = lastCommand;
		} else {
			BetterShopLogger.Warning("Stranger Attempted to modify last command: " + stm[1].toString());
		}
	}

	public static Plugin getPlugin() {
		return bettershopPlugin;
	}

	public static BSConfig getSettings() {
		return config;
	}

	public static String getLastCommand() {
		return lastCommand;
	}

	public static BSPriceList getMainPricelist() {
		return shopManager.getShop((String) null).pricelist;
	}

	public static BSPriceList getPricelist(Location l) {
		return shopManager.getShop(l).pricelist;
	}

	public static BSPriceList getPricelist(CommandSender player) {
		return getShop(player).pricelist;
	}

	public static BSSignShop getSignShop() {
		return signShop;
	}

	public static BSChestShop getChestShop() {
		return chestShop;
	}

	public static BSItemStock getStock(Location l) {
		return shopManager.getShop(l).stock;
	}

	public static BSItemStock getStock(CommandSender player) {
		return getShop(player).stock;
	}

	public static void checkRestock() {
		shopManager.checkRestock();
	}

	public static void restock() {
		shopManager.restock();
	}

	public static BSTransactionLog getTransactions(Location l) {
		return shopManager.getShop(l).transactions;
	}

	public static BSTransactionLog getTransactions(CommandSender player) {
		return getShop(player).transactions;
	}

	public static Shop getMainShop() {
		return shopManager.getShop((String) null);
	}

	public static Shop getShop(Location l) {
		return shopManager.getShop(l);
	}

	public static Shop getShop(String s) {
		return shopManager.getShop(s);
	}

	public static Shop getShop(CommandSender player) {
		return shopManager.getShop(player instanceof Player ? ((Player) player).getLocation() : null);
	}

	public static Collection<Shop> getShops() {
		return shopManager.getShops();
	}

	public static RegionShopManager getShopManager() {
		return shopManager;
	}

	public static boolean commandShopEnabled(Location loc) {
		return shopManager.isCommandShopEnabled(loc);
	}

	public static boolean spoutEnabled(){
		return buttonListener != null;
	}
	
	public static TransactionHandler.TransactionResult executeTransaction(Player player, PlayerTransation action, ShopMethod method) {
		return shopHandler.execute(player, action, method);
	}

	public static int reload(CommandSender sender) {
		int errors = 0;
		if (config.signShopEnabled && signShop.saveDelayActive()) {
			if (signShop.save()) {
				BSutils.sendMessage(sender, BetterShop.getSignShop().numSigns() + " signs saved.");
			} else {
				BSutils.sendMessage(sender, ChatColor.RED + "shop signs db save error");
				++errors;
			}
		}

		if (config.signShopEnabled && config.tntSignDestroyProtection) {
			signShop.stopProtecting();
		}
		if (JItemDB.load(BSConfig.itemDBFile)) {
			BSutils.sendMessage(sender, JItemDB.size() + " items loaded.");
		} else {
			BetterShopLogger.Warning("Cannot Load Items db!");
			BSutils.sendMessage(sender, ChatColor.RED + "Item Database Load Error.");
			++errors;
		}

		if (!BetterShop.getSettings().load()) {
			BSutils.sendMessage(sender, ChatColor.RED + "Config loading error.");
			++errors;
		} else {
			BSutils.sendMessage(sender, "Config.yml loaded.");
		}

		shopManager.closeAll();
		errors += shopManager.load();

		if (config.signShopEnabled && signShop.load()) {
			BSutils.sendMessage(sender, BetterShop.getSignShop().numSigns() + " signs loaded.");
		} else {
			BSutils.sendMessage(sender, ChatColor.RED + "shop signs db load error");
			++errors;
		}
		if (BetterShop.getSettings().signShopEnabled && BetterShop.getSettings().tntSignDestroyProtection) {
			BetterShop.getSignShop().startProtecting();
		}
		return errors;
	}

	public static class ServerReload extends TimerTask {

		Server reload = null;

		public ServerReload(Server s) {
			reload = s;
		}

		public void start(long wait) {
			(new Timer()).schedule(this, wait);
		}

		@Override
		public void run() {
			if (reload != null) {
				reload.reload();
			}
		}
	}
}
